/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

 
//                          ttt version: 1.0


//  xxxxxxxxxxxxxxx    global variable declarations  xxxxxxxxxxxxxxxx

  
// These variables below can be directly modified by the user on the options form: 

var t_time = 60000;  // task time in miliseconds (ms) 
 
var gtype = 1;       // generator type. Range: [1..3] 

var radius = 6;	    // target radius in percents of canvas width/height. units: percents converted into pixels (px) in setPar()

var speed = 5;       // target speed. Higher value = targets move faster and vice versa.

var c_time = 10000;  // change (refresh) time for stimulus type in miliseconds (ms). callback function draw() will change stimulus type approx at this time point
 
var demo = true;    // if (demo == true) => Show feedback to user (red or blue target color) 

var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 

// Other global constants and variables:

var is_task = false;    // task in progress or not. is_task is true for all events between startTask() and theEnd() events

var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_storage = false; // local storage is enabled on user browser or not

var data = {};     // application data for target animation (object)

var shift = 1;    //  basic target shift units: 1 px. updated in setPar()

var dv = 0;				// dependent variable during task (result)

var cnt = 0;    	// global counter (increase in draw/animation functions)

var xMouse, yMouse;  // mouse pointer coordinates (px) 

var errs = [0, 0, 0, 0, 0, 0, 0];  // errors array. index 0 => page was resized during task. index 1..index6 => target was out of visible canvas range for this data.type = [1..6]  
		
var startTime;    // start time for task. init in first draw() callback

var start;  		  // time variable for refresh (change) stimulus type. see draw() callback


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */


// application init [check storage, show start panel..] 
function init() 
{     
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }				  
	setIns();		
	showStart(); 
}


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// raised when page is resized. see html: <body onresize="pageResize(true || false)"
// strech canvas layout
function pageResize(par)  
{ 
	if (is_task || par)  { 
		let c= document.getElementById('can');
		c.width= window.innerWidth;  
		c.style.width= c.width + 'px';	     
		c.height= window.innerHeight;   
		c.style.height= c.height + 'px';  // https://css-tricks.com/whats-the-difference-between-width-height-in-css-and-width-height-html-attributes/ 							
		if (!par) { 
		  errs[0]++;
    } 		 
	}	
}	


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('hide_option').style= 'text-align:center';  // bug? auto-remove style when element.style.display="none"
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (is_storage && localStorage.getItem('ttt-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('ttt-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';  // Watch this line: this was a nasty bug 
		}
	}
  dirty= false;
  //document.getElementById('is_valid').textContent='';  // tmp reset validator message
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').classList.add('center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';          
  document.getElementById('hide_option').style.display = 'none';   
  //document.getElementById('start_div').style= 'text-align:center';  // bug?  (firefox auto remove class "center" when set style.display.none;)
  document.getElementById('show_option').style= 'text-align:center';  
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
  setIns();	
}


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';
  document.getElementById('ins_time').textContent= t_time/(60*1000);  // Math.round ( t_time/(60*1000) ).toString();	
  document.getElementById('ins_demo').style.display= demo ? 'block' : 'none';		
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{
		
		case 't_time':
			if (e.validity.valid)  {
				msg.textContent= 'Total task time (units: minutes)'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break;  				
			
    case 'gtype':                                                                                                                                            
      if (e.value == '1')  {
        msg.textContent= 'The Random generator uses 2 simple movements: moving along the arc and moving along the line. As soon as it reaches the end of the arc/line, a new random line/arc is generated.';
      }
      else if (e.value == '2')  {
        msg.textContent= 'The Repetitive generator uses 4 movements: sine function, movement inside the same arc, movement like a billiard ball, movement within the same line.';
      }
      else  {
        msg.textContent= 'The Mixed generator uses all possible trajectories/paths. Mixed = Random + Repetitive.'; 
      }
      break;    
			        
    case 'radius':                                                                                                                                     
      msg.textContent= e.options[e.selectedIndex].value + '%';
      break;    							
	
		case 'speed':
			if (e.validity.valid)  {
				msg.textContent= 'The speed of the moving target (circle) on the screen. Set a higher value for faster target movement and vice versa.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
			break; 
		
		case 'c_time':
			msg.textContent= 'The type of movement (path, trajectory) of the target will change approximately every ' +  e.options[e.selectedIndex].textContent;
			break;
		
		case 'demo': 
       msg.textContent= e.checked ? 'On' : 'Off';  
       break; 	
			
		case 'store':                                                                                                                                            
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty === true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// check for validity.valid object in all input type="number"
function optionsValid()  
{ 
  let res= document.getElementById('t_time').validity.valid && document.getElementById('speed').validity.valid; 	
  return res;              
}


// start task (research) 
function startTask(e)   
{       			
	initLayout();
  
	pageResize(true);	  	// must be called before startTask () => window.radius
	
	setPar();
	
  initData(null);	  
		
	is_task= true;     //  see draw(), pageResize()
	
	window.requestAnimationFrame(draw);	// activate callback Window.requestAnimationFrame() 
			
}


// init visible and invisible html dom elements before start task
function initLayout()
{
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';               	  
	document.getElementById('can').style.display= 'block';			    
	document.getElementById('can').getContext('2d').fillColor= '#000';  // default color is black	 
}

 
// set internal application research parameters (global vars)
function setPar()  
{  		 		
	cnt= 0; 
	dv= 0;	 
	errs= [0, 0, 0, 0, 0, 0, 0];  // index 0 => page was resized during task. index 1..index6 => target was out of visible canvas range for this data.type. type = [1..6]
  if ( document.getElementById('radius').selectedIndex > -1 ) {  // update window.radius. convert percents into pixels
		let r = parseInt ( document.getElementById('radius').value, 10 );
		radius = Math.floor ( r * Math.min( document.getElementById('can').width, document.getElementById('can').height ) / 100 );
	}	
	shift= speed * 1;  // basic unit is 1 px (change this formula if you want) 
	startTime= -1;     // init start time (see draw callback)
}


/*  init animation data. par: m_type => requested movement type
		switch(m_type == data.type)
    case null : random (movement type depends of gtype)
		case 1:  // Bounce off the walls (a ball bouncing off walls, like a billiard ball) 
		case 2:  // moving to the end of line AND THEN SINUS ANIMATION
		case 3:	 // moving always within same line (with repeat)
		case 4:  // moving within arc (with repeat)		
		case 5:	 // moving to the end of line (no repeat)
		case 6:  // moving within arc (no repeat)
		data.type 1..4 are repeated
		data.type 5..6 are not repeated
*/
function initData(m_type)
{				  
	let c= document.getElementById('can'), w = c.width, h = c.height, kind;
	
	if ( m_type > 0 )  {   
		kind = m_type;  // generate (init) requested movement type
	}
	else  		// m_type == null
	{
		if ( gtype == 1 ) {  // random movement
			kind = data.type || getRndInt(5,6);   // only 5 or 6 
			kind= ( kind == 5 ) ? 6 : 5; 
		}
		else if ( gtype == 2 )  {  // repetitive movement
			kind = data.type || getRndInt(1,4);
			kind++;
			if ( kind > 4 )  {
				kind = 1;      // posible range for data.type = 1..4
			}					 
		}
		else  {  // all movements => mixed generator = random + repetitive movement
			kind = data.type || getRndInt(1,6);
			kind++;
			if ( kind > 6 )  {
				kind = 1;      // posible range for data.type = 1..6
			}					 
		}			
	}  
  
	let x= data.x ? data.x : getRndInt(radius, w - radius);
	let y= data.y ? data.y : getRndInt(radius, h - radius);

	data= { type: kind, x: x, y: y };  // always exists
  
	let x2, y2, d;  
	switch (kind)
	{			  
		
		case 1:  // Bounce off the walls (a ball bouncing off walls, like a billiard ball) 
				data.dx= shift;
				data.dy= -shift;			
			break;
		
		case 2:  // moving to the end of line AND THEN SINUS ANIMATION		
			let hor=  Math.random() < 0.5 ? 1 : 2;	
			if ( hor == 1 )  
			{  // horizontal sinus function
					//do  {
					//y2= Math.random() < 0.5 ? 0.25 * h + radius : 0.75 * h - radius; => borderline for debug
				x2= getRndInt(radius, w-radius);     //x2= getRndInt(Math.round(w/4), Math.round(3*w/4));							
				y2= getRndInt( Math.ceil(0.25 * h + radius), Math.floor(0.75 * h - radius) );			
					//} while ( Math.abs(y2 - h/2) < 0.2 * h );	// 2 * amplitude = minimum 20% canvas height 
			}
			else  
			{ // vertical sinus function
					//do  {		
					//x2= Math.random() < 0.5 ? Math.ceil(0.25 * w + radius) : Math.floor(0.75 * w - radius); // debug					
				x2= getRndInt( Math.ceil(0.25 * w + radius), Math.floor(0.75 * w - radius) );  										
				y2= getRndInt(radius, h-radius);  //getRndInt(Math.round(h/4), Math.round(3*h/4));				
					//} while ( Math.abs(x2 - w/2) < 0.2 * w );	// 2 * amplitude = minimum 20% canvas width
			}	
			d = getDistance(x, y, x2, y2);		
			data.dir = hor; // data.dir == 1 => horizontal animation (graphs). data.hor == 2 =>  vertical animation (graphs) of the Sine function
			data.done = 0;	// only for data.type = 2	        									
			break;			
		
		case 3:	// moving always within same line (with repeat)
		case 5:	// moving to the end of line (no repeat)
			do  {
				x2= getRndInt(radius, w-radius); //getRndInt(w/4, 3*w/4);
				y2= getRndInt(radius, h-radius); //getRndInt(h/4, 3*h/4); 
				d = getDistance(x, y, x2, y2);
			} while ( d < 0.25 * Math.min(w,h) );		
			break;
		
		case 4:  // moving within arc (with repeat) 				
		case 6:  // moving within arc (no repeat) 
			do  {
				x2= getRndInt(radius, w-radius); // getRndInt(w/4, 3*w/4);
				y2= getRndInt(radius, h-radius); // getRndInt(h/4, 3*h/4);  
				d = getDistance(x, y, x2, y2);
			} while ( d < 0.25 * Math.min(w,h) );	
			let v = Math.sqrt(3)/2 * d; // equilateral triangle height from any side. equilateral triangle is a triangle in which all three sides have the same length. 
			let alpha = Math.atan2(y2-y, x2-x) + Math.PI/2;
			let xc = (x + x2)/2 + v * Math.cos(alpha);  
			let yc = (y + y2)/2 + v * Math.sin(alpha);											
			// https://stackoverflow.com/questions/45651736/how-can-i-draw-arcs-in-angles-of-an-arbitrary-triangle					
			let dx1 = x - xc, dy1 = y - yc,	dx2 = x2 - xc, dy2 = y2 - yc;   // get vectors from xc,yc to the other points
			let angle1 = Math.atan2(dy1,dx1), angle2 = Math.atan2(dy2,dx2);
			// use cross product to find correct direction.
			if ( dx1 * dy2 - dy1 * dx2 < 0)  {  // wrong way around swap direction
					let t = angle2;  // swap angles
					angle2 = angle1;
					angle1 = t;
			}										
			if ( angle1 < angle2 )  {		// if angle 1 is behind then move ahead
				angle1 += Math.PI * 2;
			}																		
			data.xc = xc;
			data.yc = yc;
			data.angle = angle1;
			//data.angle1 = angle1;  // debug
			//data.angle2 = angle2;  // debug
			data.n = 0;
			data.r = d;
			data.inc = 0.1 * speed * Math.PI / 180;   	// you can change this formula (speed of movement for arc animation)	  			
			break;				
		
	}   // switch kind

  if ( kind >= 2 && kind <= 6 )  
	{  					
		data.x1 = x;         
		data.y1 = y; 
		data.x2 = x2;
		data.y2 = y2;				
		if ( kind != 4 && kind != 6 )  {  // if ( typeof data.angle === 'undefined' && typeof data.inc === 'undefined' && typeof data.r === 'undefined' )
			data.angle = Math.atan2(y2-y, x2-x);
			data.r = radius;
			data.r_max = Math.floor(d - radius);
			data.inc = shift;		
		}
	}
  
}


// init horizontal (or vertical) animation (graph) of the Sine function
function initSin()
{
	data.done = 1;
	let c = document.getElementById('can'), w = c.width, h = c.height;	  
	data.N = getRndInt(1,3);	 // number of sine cycles within canvas region bounds			     	
	let hor = data.dir == 1 ? true: false;  // horizontal movement or vertical  
	if ( hor == 1 )
	{
		// x
		data.x1 = data.x;     // data.x1= data.x < radius || data.x > w-radius ? 2*radius : data.x;
		data.x2 = data.x <= w/2 ? w - radius : radius;  // edge of canvas object			
		data.frequency = Math.floor( ( Math.abs(data.x2 - data.x1) - radius ) / (2 * data.N * Math.PI));
		// y
		data.baseline = data.y; 			
		let y_off= Math.abs(h/2 - data.y) - radius;			
		data.y1 = data.y - y_off;
		data.y2 = data.y + y_off;       																
		data.amplitude = y_off - radius;  // this can be very small value. you should change initData() to limit amplitude
		data.inc = data.x1 <= data.x2 ? shift : -shift;					 		
	}	
	else
	{
		// y
		data.y1 = data.y;  
		data.y2 = data.y <= h/2 ? h - radius : radius;  // edge of canvas object			
		data.frequency = Math.floor( ( Math.abs(data.y2 - data.y1) - radius ) / (2 * data.N * Math.PI));
		// x
		data.baseline = data.x; 			
		let x_off= Math.abs( w/2 - data.x) - radius;			
		data.x1 = data.x - x_off;
		data.x2 = data.x + x_off;       																
		data.amplitude = x_off - radius;   // this can be very small value. you should change initData() to limit amplitude
		data.inc = data.y1 <= data.y2 ? shift : -shift;		
	}			
}


// return random integer in range [min..max]
function getRndInt(min, max) {
		return Math.floor(Math.random() * (max - min + 1) ) + min;
}


/*
https://abhi9bakshi.medium.com/why-javascript-timer-is-unreliable-and-how-can-you-fix-it-9ff5e6d34ee0
https://developer.mozilla.org/en-US/docs/Web/API/window/requestAnimationFrame
https://html.spec.whatwg.org/multipage/imagebitmap-and-animations.html#dom-animationframeprovider-requestanimationframe
https://w3c.github.io/hr-time/#dom-performance-now
A callback is a function passed as an argument to another function.  */
		/*  switch(data.type)
		case 1:  // Bounce off the walls (a ball bouncing off walls, like a billiard ball) 
		case 2:  // moving to the end of line AND THEN SINUS ANIMATION
		case 3:	 // moving always within same line (with repeat)
		case 4:  // moving within arc (with repeat)		
		case 5:	 // moving to the end of line (no repeat)
		case 6:  // moving within arc (no repeat)
		data.type 1..4 are repeated
		data.type 5..6 are not repeated
*/		
function draw(timestamp) 
{
	if ( is_task )
	{

		if (startTime == -1)  {  
			startTime = timestamp;	// initialize time when task was started
			start = timestamp;	 	  // initialize time for change/refresh stimulus type		
		}
		
		let elapsed = timestamp - start;  // const elapsed = Math.round(timestamp - start); 		
		if ( Math.round(timestamp - startTime) > t_time)  {  // is it end of task?
			is_task= false;   
			theEnd ( Math.round((timestamp - startTime)/1000) );  // seconds (convert time difference from milliseconds to seconds)
			return;
		}
		else if (elapsed >= c_time)  { 	// change the animation type after each x seconds			  			
			initData(null);
			start = timestamp;				
		}
					
		switch (data.type)
		{
			case 1:   // Bounce off the walls (a ball bouncing off walls, like a billiard ball)
				animBall();
				break;			
				
			case 2:  // moving to the end of line AND THEN SINUS ANIMATION	
				if (data.done == 0 ) {
					animLine();
				}
				else  {
					animSin();
				}
				break;
				
			case 3:	// moving always within same line (with repeat)
			case 5:	// moving to the end of line (no repeat)
				animLine();
				break;	
				
			case 4:  // moving within arc (with repeat) 				
			case 6:  // moving within arc (no repeat) 
				animArc();
				break;	
			
		}
		
		// Note: Your callback routine must itself call requestAnimationFrame() again if you want to animate another frame at the next repaint. requestAnimationFrame() is 1 shot.				
		window.requestAnimationFrame(draw);  	
	}
}


// https://developer.mozilla.org/en-US/docs/Games/Tutorials/2D_Breakout_game_pure_JavaScript/Bounce_off_the_walls
function animBall() 
{
	let c = document.getElementById('can'), ctx = c.getContext('2d'), w = c.width, h = c.height;					
	ctx.clearRect(0, 0, w, h);
	
	if ( data.x + data.dx < radius || data.x + data.dx > w - radius ) {
		data.dx = -data.dx;
	}
	if ( data.y + data.dy < radius || data.y + data.dy > h - radius ) {
		data.dy = -data.dy;
	}					
	data.x += data.dx;
	data.y += data.dy;

	let ok= getDistance(xMouse, yMouse, data.x, data.y) <= radius ? true: false;
	if (demo)  {
		ctx.fillStyle= ok ? '#00f' : '#f00'; 
	}	
	ctx.beginPath();
	ctx.arc(data.x, data.y, radius, 0, Math.PI*2);
	ctx.fill();
	
  cnt++;
	if ( ok ) { 
		dv++;
	}	
	if (data.x < radius || data.x > w - radius || data.y < radius || data.y > h - radius ) {					
		errs[data.type]++;		  
	}	
}		
		
		
// draws horizontal or vertical animation (virtual graph) of the Sine function
function animSin()
{ 
	
	let c = document.getElementById('can'), ctx = c.getContext('2d'), w = c.width, h = c.height;			
	ctx.clearRect(0, 0, w, h);     // clear canvas
		 	
	let hor = data.dir == 1 ? true: false;  // horizontal movement or vertical => data.dir == 1 => horizontal animation (graphs). data.hor == 2 =>  vertical animation (graphs) of the Sine (or Cosine) Function 
	ctx.beginPath();
	if ( hor == 1 )  
	{
		if ( data.x + data.inc < Math.min(data.x1, data.x2) || data.x + data.inc > Math.max(data.x1,data.x2) ) {
			data.inc = -data.inc;
		}			
		data.x += data.inc;					 
		data.y = data.baseline + data.amplitude * Math.sin(data.x/data.frequency);			
	}	
	else 
	{
		if ( data.y + data.inc < Math.min(data.y1, data.y2) || data.y + data.inc > Math.max(data.y1,data.y2) ) {
			data.inc = -data.inc;
		}			
		data.y += data.inc;	
		data.x = data.baseline + data.amplitude * Math.sin(data.y/data.frequency);												
	}		
  
	ctx.arc(data.x, data.y, radius, 0, 2 * Math.PI); 
	
	let ok= getDistance(xMouse, yMouse, data.x, data.y) <= radius ? true: false; 
	if ( demo )  {
		ctx.fillStyle = ok ? '#00f' : '#f00';
	}
	ctx.fill();	
	
	cnt++;
	if ( ok ) { 
		dv++;
	}
	
	if (data.x < radius || data.x > w - radius || data.y < radius || data.y > h - radius ) {					
		errs[data.type]++;		  
	}		
	
}


// handle line animation for data.type = 3, data.type= 5 and first part for data.type= 2
function animLine() 
{
	let c = document.getElementById('can'), ctx = c.getContext('2d'), w = c.width, h = c.height;					
	ctx.clearRect(0, 0, w, h);
	let overflow= data.r + data.inc < radius || data.r + data.inc > data.r_max; 
	if ( overflow )  {			  
		if (data.type == 2 && data.done == 0){
			initSin();
			return;
		}
		else if (data.type == 3) {
			data.inc = -data.inc;
		}
		else  /* if (data.type == 5) */ {
			initData(5); 
			return;
		}
	}						
	data.r += data.inc;	
	data.x = data.x1 + data.r * Math.cos(data.angle);  // render the line using data.r pixel radius - x
	data.y = data.y1 + data.r * Math.sin(data.angle);  // render the line using data.r pixel radius - y
	
	let ok = getDistance(xMouse, yMouse, data.x, data.y) <= radius ? true: false; 
	if ( demo )  {
		ctx.fillStyle = ok ? '#00f' : '#f00';
	}
	ctx.beginPath();
	ctx.arc(data.x, data.y, radius, 0, Math.PI*2);
	ctx.fill();

  cnt++;
	if ( ok ) { 
		dv++;
	}
	
	if (data.x < radius || data.x > w - radius || data.y < radius || data.y > h - radius ) {					
		errs[data.type]++;		  
	}	
}		


/* moving within arc from point P1 to P2 
	 for data.type == 4 => with repeat (overflow)
	 for data.type == 6 => no repeat  */
function animArc() 
{
	let c = document.getElementById('can'), ctx = c.getContext('2d'), w = c.width, h = c.height;											
	data.n+= data.inc;  
	data.angle += data.inc; 
	data.x = data.xc + data.r * Math.cos(data.angle);  // render the line using center in point P(data.xc,data.yc) 
	data.y = data.yc + data.r * Math.sin(data.angle);  // and data.r pixel radius and data.angle in radians
	
	if ( data.x < radius || data.y < radius || data.x > w - radius || data.y > h - radius || ( data.type == 6 && data.n > Math.PI/3) ) {
		initData(6);   // generate new virtual random equilateral triangle (with 3 angles= Math.PI/3)
		return;     				
	}		     
	else if ( data.type == 4 && Math.abs(data.n) > Math.PI/3)  {  // with repeat
		data.inc = -data.inc;
		data.n= 0; 
		return;
	}
		
	ctx.clearRect(0, 0, w, h);			
	
	let ok= getDistance(xMouse, yMouse, data.x, data.y) <= radius ? true: false; 
	if ( demo )  {
		ctx.fillStyle = ok ? '#00f' : '#f00';
	}
	ctx.beginPath();
	ctx.arc(data.x, data.y, radius, 0, Math.PI*2);
	ctx.fill();

  cnt++;
	if ( ok ) { 
		dv++;
	}
	
}		


/* html canvas => mousemove event	=> a pointing device is moved over an element (fired continuously as the mouse moves).
   this event is fired by html canvas element => document.getElementById('can') 
	 <canvas id="can" onmousemove="mouseMove(event)"></canvas>  */
function mouseMove(e)
{  
	// remember the last coordinates   				
	if (is_task) {
		xMouse= e.clientX;  
		yMouse= e.clientY; 
	} 	
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
	
	if (document.getElementById('t_time').validity.valid) {
		t_time= 60 * 1000 * parseInt(document.getElementById('t_time').value, 10);  // task time (in milliseconds)
  }	
	
	if ( document.getElementById('gtype').selectedIndex > -1 ) {
		gtype= parseInt(document.getElementById('gtype').value, 10); 
	}
	
	// window.radius is set latter in setPar()
	
	if (document.getElementById('speed').validity.valid) {
	  speed= parseInt(document.getElementById('speed').value, 10); 
  }	

  if ( document.getElementById('c_time').selectedIndex > -1 ) {
		c_time= parseInt(document.getElementById('c_time').value, 10); 
	}

  demo= document.getElementById('demo').checked ? true: false;	
	
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false;
  	
}


// end of task (research)
// par: task_time => converted task time difference from milliseconds to seconds. this timestamp difference is set in callback draw()
function theEnd(task_time)  
{     
  document.getElementById('can').style.display= 'none'; 
  document.getElementById('end').style.display= 'block';   
  document.getElementById('result_wrapper').style.display= 'block'; 
  
	if ( errs[0] > 0 ) { 
		document.getElementById('err_r').style.display= 'block'; 
		document.getElementById('err_r').textContent= 'Warning: window (page) was resized during task (error-r= ' + errs[0] + ')';
	}
	else  {
		document.getElementById('err_r').style.display= 'none';
	}
	let err_i= 0;
	for (let i= 1; i < errs.length; i++) {
		err_i += errs[i]; 
	} 
	if ( err_i > 0 )  {
		document.getElementById('err_i').style.display= 'block';
		document.getElementById('err_i').textContent= 'Warning: target was out of visible screen bounds - partially or completely (error-i= ' + err_i + ')';
	}
	else  {
		document.getElementById('err_i').style.display= 'none';
	}	
	
	let ok= (100*dv/cnt), not_ok= 100 - ok;   // ok = dependent variable converted into percents
  ok= ok.toFixed(2);  // convert number (percents) into string	   	
  not_ok = not_ok.toFixed(2); // convert number (percents) into string	
	
	// "chart" => repaint "chart" 	
  document.getElementById('chart_1').setAttribute('style', 'width:' + ok + '%');
	document.getElementById('chart_2').setAttribute('style', 'width:' + not_ok + '%');    // !important	
	document.getElementById('chart').title= 'Performance (result) in percentages= ' + ok + '%';  	
		
	createTableArray(document.getElementById('table1'), [ok, not_ok],  ['Successfully (%)', 'Unsuccessful (%)']); 
	 		
	document.getElementById('table1').style.display= 'block';	
		
	// update 'score' label => above "Repeat" button
	document.getElementById('score').textContent= 'Percentage of successful target tracking. Result= ' + ok + '%.'; 
	
  document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
	document.getElementById('table_grp').style.display= 'none';
	try 
  {
    if ( store ) {
			setStorage(task_time, err_i, ok);  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }    	 
 
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if ( mm < 10 ) { mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:
	 time_s => measured task time in seconds by callback function draw()
	 err_i => the target was partially out of the visible part of the screen - how many times this happened 
	 ok: window.dv string in percentage with fixed decimals (task result)		*/  		
function setStorage(time_s, err_i, ok)   
{    
	let oldItems = JSON.parse(localStorage.getItem('ttt-v1.0')) || []; 
  let id= oldItems.length+1;     	 
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   // new Date(),  		   
      'time-m' : Math.round( t_time / (60*1000) ), 	  // convert ms into minutes
			'generator' : gtype,
      'size': document.getElementById('radius').value, 
			'speed': speed,
			'change-s': Math.round(c_time/1000),   // convert ms into seconds
			'demo' : demo ? 1: 0,
      'result':  ok,			
			'time-s' : time_s,
			'error-r': errs[0],   // err[0] => number of pageResize() event when is_task == true 
      'error-i': err_i			
  };

  oldItems.push(newItem);

  localStorage.setItem('ttt-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);
  
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement("table");  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement("th");      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined){
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('ttt-v1.0')  ?  JSON.parse(localStorage.getItem('ttt-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved application results:  '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 { 
   window.localStorage.removeItem('ttt-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// return distance from point P1(x1,y1) to point P2(x2,y2)
function getDistance(x1, y1, x2, y2)  {
  return Math.sqrt( Math.pow(x1-x2, 2) + Math.pow(y1-y2, 2) ); // total line distance (px) 
}	


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('ttt-v1.0');
	if (!data)  {
    alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// https://stackoverflow.com/questions/60698115/how-to-copy-a-javascript-2d-array-to-clipboard-to-paste-it-in-excel
// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// source from: https://stackoverflow.com/questions/400212/how-do-i-copy-to-the-clipboard-in-javascript
// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('ttt-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved cumulative data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// https://stackoverflow.com/questions/2004555/how-can-i-make-javascript-make-produce-new-page
// displays Latest research results in new empty web page (tab) 
// parameter: id => DOM/html object (div or table)
function newWinLast(id)
{ 
  let e = document.getElementById(id);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{     
  let tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// https://stackoverflow.com/questions/3065342/how-do-i-iterate-through-table-rows-and-cells-in-javascript
// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
